@@ -1,7 +1,6 @@
 package Tcl;
 
-$Tcl::VERSION = '1.00';
-$Tcl::STACK_TRACE = 1;
+$Tcl::VERSION = '1.02_50';
 
 =head1 NAME
 
@@ -371,20 +370,6 @@ to use these libraries in a link stage.
 
 =back
 
-=head1 AUTHORS
-
-Malcolm Beattie, mbeattie@sable.ox.ac.uk, 23 Oct 1994.
-Vadim Konovalov, vkon@cpan.org, 19 May 2003.
-Jeff Hobbs, jeff (a) activestate . com, 22 Mar 2004.
-Gisle Aas, gisle (a) activestate . com, 14 Apr 2004.
-
-=head1 COPYRIGHT
-
-This program is free software; you can redistribute it and/or modify it under
-the same terms as Perl itself.
-
-See http://www.perl.com/perl/misc/Artistic.html
-
 =cut
 
 use strict;
@@ -428,31 +413,21 @@ sub new {
     return $int;
 }
 
-eval {
-    require "Tclaux.pm";
-};
-
 END {
     Tcl::_Finalize();
 }
 
-#TODO make better wording here
-# %anon_refs keeps track of anonymous subroutines that were created with
-# "CreateCommand" method during process of transformation of arguments for
-# "call" and other stuff such as scalar refs and so on.
-# (TODO -- find out how to check for refcounting and proper releasing of
-# resources)
+# %anon_refs keeps track of anonymous subroutines and scalar/array/hash
+# references which are created on the fly for tcl/tk interchange
+# at a step when 'call' interpreter method prepares its arguments for
+# tcl/tk call, which is invoked by 'icall' interpreter method
+# (this argument transformation is done with "CreateCommand" method for
+# subs and with 'tie' for other)
 
 my %anon_refs;
 
-# %widget_refs is an array to hold refs that were created when working with
-# widget the point is - it's not dangerous to delete more than needed, because
-# those will be re-created at the very next time they needed.
-# however when widget goes away, it is good to delete anything that comes
-# into mind with that widget
-my %widget_refs;
-my $current_widget = '';
-sub _current_refs_widget {$current_widget=shift}
+# (TODO -- find out how to check for refcounting and proper releasing of
+# resources)
 
 # Subroutine "call" preprocess the arguments for special cases
 # and then calls "icall" (implemented in Tcl.xs), which invokes
@@ -460,23 +435,19 @@ sub _current_refs_widget {$current_widget=shift}
 sub call {
     my $interp = shift;
     my @args = @_;
+    my $current_r = join ' ', grep {defined} grep {!ref} @args;
+    my @codes;
 
     # Process arguments looking for special cases
     for (my $argcnt=0; $argcnt<=$#args; $argcnt++) {
 	my $arg = $args[$argcnt];
 	my $ref = ref($arg);
 	next unless $ref;
-	if ($ref eq 'CODE') {
+	if ($ref eq 'CODE' || $ref eq 'Tcl::Code') {
 	    # We have been passed something like \&subroutine
 	    # Create a proc in Tcl that invokes this subroutine (no args)
-	    $args[$argcnt] = $interp->create_tcl_sub($arg);
-	    $widget_refs{$current_widget}->{$args[$argcnt]}++;
-	}
-	elsif ($ref =~ /^Tcl::Tk::Widget\b/) {
-	    # We have been passed a widget reference.
-	    # Convert to its Tk pathname (eg, .top1.fr1.btn2)
-	    $args[$argcnt] = $arg->path;
-	    $current_widget = $args[$argcnt] if $argcnt==0;
+	    $args[$argcnt] = $interp->create_tcl_sub($arg, undef, undef, $current_r);
+	    push @codes, $anon_refs{$current_r}; # push CODE also only to keep it from early disposal
 	}
 	elsif ($ref eq 'SCALAR') {
 	    # We have been passed something like \$scalar
@@ -486,9 +457,7 @@ sub call {
 	    # This will be SCALAR(0xXXXXXX) - leave it to become part of a
 	    # Tcl array.
 	    my $nm = "::perl::$arg";
-	    #$nm =~ s/\W/_/g; # remove () from stringified name
 	    unless (exists $anon_refs{$nm}) {
-		$widget_refs{$current_widget}->{$nm}++;
 		$anon_refs{$nm} = $arg;
 		my $s = $$arg;
 		tie $$arg, 'Tcl::Var', $interp, $nm;
@@ -508,7 +477,6 @@ sub call {
 	    $nm =~ s/\W/_/g; # remove () from stringified name
 	    $nm = "::perl::$nm";
 	    unless (exists $anon_refs{$nm}) {
-		$widget_refs{$current_widget}->{$nm}++;
 		$anon_refs{$nm} = $arg;
 		my %s = %$arg;
 		tie %$arg, 'Tcl::Var', $interp, $nm;
@@ -530,27 +498,10 @@ sub call {
 	    $args[$argcnt] =
 		$interp->create_tcl_sub(sub {
 		    $arg->[0]->(@_, @$arg[1..$#$arg]);
-		}, $events);
+		}, $events, undef, $current_r);
+	    push @codes, $anon_refs{$current_r};
 	}
-	elsif ($ref eq 'ARRAY' && ref($arg->[0]) =~ /^Tcl::Tk::Widget\b/) {
-	    # We have been passed [$Tcl_Tk_widget, 'method name', ...]
-	    # Create a proc in Tcl that invokes said method with args
-	    my $events;
-	    # Look for Tcl::Ev objects as the first arg - these must be
-	    # passed through for Tcl to evaluate.  Used primarily for %-subs
-	    # This could check for any arg ref being Tcl::Ev obj, but it
-	    # currently doesn't.
-	    if ($#$arg >= 1 && ref($arg->[1]) eq 'Tcl::Ev') {
-		$events = splice(@$arg, 1, 1);
-	    }
-	    my $wid = $arg->[0];
-	    my $method_name = $arg->[1];
-	    $args[$argcnt] =
-		$interp->create_tcl_sub(sub {
-		    $wid->$method_name(@$arg[2..$#$arg]);
-		}, $events);
-	}
-	elsif (ref($arg) eq 'REF' and ref($$arg) eq 'SCALAR') {
+	elsif ($ref eq 'REF' and ref($$arg) eq 'SCALAR') {
 	    # this is a very special shortcut: if we see construct like \\"xy"
 	    # then place proper Tcl::Ev(...) for easier access
 	    my $events = [map {"%$_"} split '', $$$arg];
@@ -560,10 +511,12 @@ sub call {
 		$args[$argcnt] =
 		    $interp->create_tcl_sub(sub {
 			$arg->[0]->(@_, @$arg[1..$#$arg]);
-		    }, $events);
+		    }, $events, undef, $current_r);
+		push @codes, $anon_refs{$current_r};
 	    }
 	    elsif (ref($args[$argcnt+1]) eq 'CODE') {
-		$args[$argcnt] = $interp->create_tcl_sub($args[$argcnt+1],$events);
+		$args[$argcnt] = $interp->create_tcl_sub($args[$argcnt+1],$events, undef, $current_r);
+		push @codes, $anon_refs{$current_r};
 	    }
 	    else {
 		warn "not CODE/ARRAY expected after description of event fields";
@@ -571,15 +524,42 @@ sub call {
 	    splice @args, $argcnt+1, 1;
 	}
     }
+
+    if ($#codes>-1 and $args[0] eq 'after') {
+	if ($args[1] =~ /^\d+$/) {
+	    my $id = $interp->icall(@args);
+	    #print STDERR "rebind for $interp;$id\n";
+	    # in 'after' methods, disposal of CODE REFs based on 'after' id
+	    # i.e based on return value of tcl call
+	    $anon_refs{"$interp;$id"} = \@codes;
+	    delete $anon_refs{$current_r};
+	    # plan deleting that entry, hence Tcl command during Tcl::Code::DESTROY
+	    # TODO - this +1000 is wrong... should
+	    #$interp->invoke('after',$args[1]+1000, "perl::Eval {Tcl::_code_dispose('$interp;$id')}");
+	    #$interp->invoke('trace','add','variable',$id,'unset',"puts {deleting $id}");
+	    #$interp->invoke('trace','add','variable',$id,'read',"puts {deleting $id}");
+	    #$interp->invoke('trace','add','variable',$id,'write',"puts {deleting $id}");
+	    #$interp->invoke('trace','add','variable',$id,'qrite',"puts {deleting $id}");
+	    return $id;
+	} elsif ($args[1] eq 'idle') {
+	    # no planned CODE REF disposal, just do as is
+	    return $interp->icall(@args);
+	}
+	# if we're here - user does something wrong, but there is nothing we worry about
+    }
+
     # Done with special var processing.  The only processing that icall
     # will do with the args is efficient conversion of SV to Tcl_Obj.
     # A SvIV will become a Tcl_IntObj, ARRAY refs will become Tcl_ListObjs,
     # and so on.  The return result from icall will do the opposite,
     # converting a Tcl_Obj to an SV.
-    if (!$Tcl::STACK_TRACE) {
-	return $interp->icall(@args);
-    }
-    elsif (wantarray) {
+
+    # we need just this:
+    #    return $interp->icall(@args);
+    # a bit of complications only to allow stack trace, i.e. in case of errors
+    # user will get error pointing to his program and not in this module.
+
+    if (wantarray) {
 	my @res;
 	eval { @res = $interp->icall(@args); };
 	if ($@) {
@@ -600,81 +580,69 @@ sub call {
     }
 }
 
-# wcall is simple wrapper to 'call' but it tries to search $res in %anon_hash
-# This implementation is temporary
-sub wcall {
-    if (wantarray) {
-	return call(@_);
-    } else {
-	my $res = call(@_);
-	if (exists $anon_refs{$res}) {
-	    return $anon_refs{$res};
-	}
-	return $res;
-    }
-}
-
-sub return_ref {
-    my $interp = shift;
-    my $rname = shift;
-    return $anon_refs{$rname};
-}
-sub delete_ref {
-    my $interp = shift;
-    my $rname = shift;
-    my $ref = delete $anon_refs{$rname};
-    if (ref($ref) eq 'CODE') {
-	$interp->DeleteCommand($rname);
-    }
-    else {
-	$interp->UnsetVar($rname); #TODO: will this delete variable in Tcl?
-	untie $$ref;
-    }
-    return $ref;
-}
-sub return_widget_refs {
-    my $interp = shift;
-    my $wpath = shift;
-    return keys %{$widget_refs{$wpath}};
-}
-sub delete_widget_refs {
-    my $interp = shift;
-    my $wpath = shift;
-    for (keys %{$widget_refs{$wpath}}) {
-	#print STDERR "del:$wpath($_)\n";
-	delete $widget_refs{$wpath}->{$_};
-	$interp->delete_ref($_);
-    }
-}
-
-# create_tcl_sub will create TCL sub that will invoke perl anonymous sub
+# create_tcl_sub will create TCL sub that will invoke perl CODE ref
 # If $events variable is specified then special processing will be
 # performed to provide needed '%' variables.
 # If $tclname is specified then procedure will have namely that name,
 # otherwise it will have machine-readable name.
 # Returns tcl script suitable for using in tcl events.
 sub create_tcl_sub {
-    my ($interp,$sub,$events,$tclname) = @_;
+    my ($interp,$sub,$events,$tclname, $rname) = @_;
     unless ($tclname) {
 	# stringify sub, becomes "CODE(0x######)" in ::perl namespace
 	$tclname = "::perl::$sub";
     }
-    unless (exists $anon_refs{$tclname}) {
-	$anon_refs{$tclname} = $sub;
-	$interp->CreateCommand($tclname, $sub, undef, undef, 1);
-    }
+
+    #print STDERR "...=$rname\n";
+    $interp->CreateCommand($tclname, $sub, undef, undef, 1);
+
+    # following line a bit more tricky than it seems to.
+    # because the whole intent of the %anon_refs hash is to have refcount
+    # of (possibly) anonymous sub that is happen to be passed,
+    # and, if passed for the same widget but arguments are same - then
+    # previous instance will be overwriten, and sub will be destroyed due
+    # to reference count, and Tcl method will also be destroyed during
+    # Tcl::Code::DESTROY
+    $anon_refs{$rname} = bless [\$sub, $interp], 'Tcl::Code';
+
     if ($events) {
 	# Add any %-substitutions to callback
 	$tclname = "$tclname " . join(' ', @{$events});
     }
     return $tclname;
 }
+
+sub _code_dispose {
+    my $k = shift;
+    #print STDERR "_code_dispose $k\n";
+    #my $int = $anon_refs{$k}->[0]->[1];
+    #my @r = $int->Eval("after info $id"); # why do not work?
+    #print STDERR "r=@r\n";
+    delete $anon_refs{$k};
+}
+
+
 sub Ev {
     my @events = @_;
     return bless \@events, "Tcl::Ev";
 }
 
 
+package Tcl::Code;
+
+# only purpose is to track CODE REFs passed to 'call' method 
+# (often these are anon subs)
+# so to bless it to this package and then catch deleting it, so 
+# to do cleaning up
+
+sub DESTROY {
+    my $rsub = $_[0]->[0];
+    my $interp = $_[0]->[1];
+    my $tclname = "::perl::$$rsub";
+    #print STDERR "CODE::DESTROY[[@_]] $tclname\n";
+    $interp->DeleteCommand($tclname) if 0 and defined $tclname;  #TODO-ed
+}
+
 package Tcl::List;
 
 use overload '""' => \&as_string,
@@ -784,5 +752,182 @@ sub DESTROY {
 #    }
 #}
 
+package Tcl;
+
+=head1 Other Tcl interpreter methods
+
+=over 2
+
+=item export_to_tcl method
+
+An interpreter method, export_to_tcl, is used to expose a number of perl
+subroutines and variables all at once into tcl/tk.
+
+B<export_to_tcl> takes a hash as arguments, which represents named parameters,
+with following allowed values:
+
+=over 4
+
+=item B<namespace> => '...'
+
+tcl namespace, where commands and variables are to
+be created, defaults to 'perl'. If '' is specified - then global
+namespace is used. A possible '::' at end is stripped.
+
+=item B<subs> => { ... }
+
+anonymous hash of subs to be created in Tcl, in the form /tcl name/ => /code ref/
+
+=item B<vars> => { ... }
+
+anonymous hash of vars to be created in Tcl, in the form /tcl name/ => /code ref/
+
+=item B<subs_from> => '...'
+
+a name of Perl namespace, from where all existing subroutines will be searched
+and Tcl command will be created for each of them.
+
+=item B<vars_from> => '...'
+
+a name of Perl namespace, from where all existing variables will be searched,
+and each such variable will be tied to Tcl.
+
+=back
+
+An example:
+
+  use strict;
+  use Tcl;
+  
+  my $int = new Tcl;
+  
+  $tcl::foo = 'qwerty';
+  $int->export_to_tcl(subs_from=>'tcl',vars_from=>'tcl');
+  
+  $int->Eval(<<'EOS');
+  package require Tk
+  
+  button .b1 -text {a fluffy button} -command perl::fluffy_sub
+  button .b2 -text {a foo button} -command perl::foo
+  entry .e -textvariable perl::foo
+  pack .b1 .b2 .e
+  focus .b2
+  
+  tkwait window .
+  EOS
+  
+  sub tcl::fluffy_sub {
+      print "Hi, I am a fluffy sub\n";
+  }
+  sub tcl::foo {
+      print "Hi, I am foo\n";
+      $tcl::foo++;
+  }
+
+=cut
+
+sub export_to_tcl {
+    my $int = shift;
+    my %args = @_;
+
+    # name of Tcl package to hold tcl commands bound to perl subroutines
+    my $tcl_namespace = (exists $args{namespace} ? $args{namespace} : 'perl::');
+    $tcl_namespace=~s/(?:::)?$/::/;
+
+    # a batch of perl subroutines which tcl counterparts should be created
+    my $subs = $args{subs} || {};
+
+    # a batch of perl variables which tcl counterparts should be created
+    my $vars = $args{vars} || {};
+
+    # TBD:
+    # only => \@list_of_names
+    # argument to be able to limit the names to export to Tcl.
+
+    if (exists $args{subs_from}) {
+	# name of Perl package, which subroutines would be bound to tcl commands
+	my $subs_from = $args{subs_from};
+	$subs_from =~ s/::$//;
+	no strict 'refs';
+	for my $name (keys %{"$subs_from\::"}) {
+	    #print STDERR "$name;\n";
+	    if (defined &{"$subs_from\::$name"}) {
+		if (exists $subs->{$name}) {
+		    next;
+		}
+		#print STDERR "binding sub '$name'\n";
+		$int->CreateCommand("$tcl_namespace$name", \&{"$subs_from\::$name"}, undef, undef, 1);
+	    }
+	}
+    }
+    if (exists $args{vars_from}) {
+	# name of Perl package, which subroutines would be bound to tcl commands
+	my $vars_from = $args{vars_from};
+	$vars_from =~ s/::$//;
+	no strict 'refs';
+	for my $name (keys %{"$vars_from\::"}) {
+	    #print STDERR "$name;\n";
+	    if (defined ${"$vars_from\::$name"}) {
+		if (exists $vars->{$name}) {
+		    next;
+		}
+		#print STDERR "binding var '$name' in '$tcl_namespace'\n";
+		local $_ = ${"$vars_from\::$name"};
+		tie ${"$vars_from\::$name"}, 'Tcl::Var', $int, "$tcl_namespace$name";
+		${"$vars_from\::$name"} = $_;
+	    } 
+	    if (0) {
+		# array, hash - no need to do anything.
+		# (or should we?)
+	    }
+	}
+    }
+
+    for my $subname (keys %$subs) {
+	#print STDERR "binding2 sub '$subname'\n";
+        $int->CreateCommand("$tcl_namespace$subname",$subs->{$subname}, undef, undef, 1);
+    }
+
+    for my $varname (keys %$vars) {
+	#print STDERR "binding2 var '$varname'\n";
+	unless (ref($vars->{$varname})) {
+	    require 'Carp.pm';
+	    Carp::croak("should pass var ref as variable bind parameter");
+	}
+	local $_ = ${$vars->{$varname}};
+	tie ${$vars->{$varname}}, 'Tcl::Var', $int, "$tcl_namespace$varname";
+	${$vars->{$varname}} = $_;
+    }
+}
+
+=item B<export_tcl_namespace>
+
+extra convenience sub, binds to tcl all subs and vars from perl B<tcl::> namespace
+
+=back
+
+=cut
+
+sub export_tcl_namespace {
+    my $int = shift;
+    $int->export_to_tcl(subs_from=>'tcl', vars_from=>'tcl');
+}
+
+=head1 AUTHORS
+
+ Malcolm Beattie, mbeattie@sable.ox.ac.uk, 23 Oct 1994.
+ Vadim Konovalov, vkon@cpan.org, 19 May 2003.
+ Jeff Hobbs, jeff (a) activestate . com, 22 Mar 2004.
+ Gisle Aas, gisle (a) activestate . com, 14 Apr 2004.
+
+=head1 COPYRIGHT
+
+This program is free software; you can redistribute it and/or modify it under
+the same terms as Perl itself.
+
+See http://www.perl.com/perl/misc/Artistic.html
+
+=cut
+
 1;
 
